;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "lsp-docker" "20260507.1750"
  "LSP Docker integration."
  '((emacs    "29.1")
    (dash     "2.14.1")
    (lsp-mode "6.2.1")
    (f        "0.20.0")
    (s        "1.13.0")
    (yaml     "0.2.0")
    (ht       "2.0"))
  :url "https://github.com/emacs-lsp/lsp-docker"
  :commit "f666fba72b496c7750bb3f349771b07aa51714f0"
  :revdesc "f666fba72b49"
  :keywords '("languages" "langserver")
  :authors '(("Ivan Yonchovski" . "yyoncho@gmail.com"))
  :maintainers '(("Ivan Yonchovski" . "yyoncho@gmail.com")))
